from typing import MutableMapping

import tinytim.data as data_features
import tinytim.utils as utils


def data_columns_same_len(data: MutableMapping) -> bool:
    """Check if data columns are all the same len."""
    if data_features.column_count(data) == 0: return True
    it = iter(data.values())
    the_len = len(next(it))
    return all(len(l) == the_len for l in it)


def valid_table_mapping(data: MutableMapping) -> bool:
    """Check if data is a true TableMapping."""
    if not utils.has_mapping_attrs(data): return False
    return data_columns_same_len(data)


from typing import Callable, Mapping, Sequence, Tuple, Any, Union, Optional, Dict, List, Iterable
from itertools import repeat
from typing import NamedTuple

class MatchIndexes(NamedTuple):
    value: Any
    left_index: Optional[int] = None
    right_index: Optional[int] = None

DataMapping = Mapping[str, Sequence]
DataDict = Dict[str, list]
Matches = Tuple[MatchIndexes]
JoinStrategy = Callable[[Sequence, Sequence], Matches]


def locate(
    l: Sequence,
    value: Any
) -> List[int]:
    """
    Return list of all index of each matching item in list.
    
    Parameters
    ----------
    l : Sequence
        sequence of values to check for matches
    value : Any
        value to check if equals any values in l
        
    Returns
    -------
    List[int]
        index numbers of items in l that equal value
        
    Example
    -------
    >>> l = [1, 2, 1, 2, 4, 5, 1]
    >>> locate(l, 1)
    [0, 2, 6]
    """
    return [i for i, x in enumerate(l) if x == value]


def _name_matches(matches: Iterable[Tuple[Any, int, int]]) -> Tuple[MatchIndexes]:
    """
    Convert tuples into MatchIndexes namedtuples.
    
    Parameters
    ----------
    matches : Iterable[Tuple[Any, int, int]]
        iterable of matches tuples[value, left_index, right_index]
        
    Returns
    -------
    Tuple[MatchIndexes]
        tuple of MatchIndexes namedtuples[Any, Optional[int], Optional[int]]
        
    Example
    -------
    >>> matches = [('a', 1, 1), ('b', 2, None), ('c', None, 3)]
    >>> name_matches(matches)
    (MatchIndexes(value='a', left_index=1, right_index=1),
     MatchIndexes(value='b', left_index=2, right_index=None),
     MatchIndexes(value='c', left_index=None, right_index=3))
    """
    return tuple(MatchIndexes(value, left_index, right_index)
        for value, left_index, right_index in matches)


def inner_matching_indexes(
    left: Sequence,
    right: Sequence
) -> Tuple[MatchIndexes]:
    """
    Find matching item value indexes in two lists.
    Returns tuple of tuples: namedtuple[value, left_index, right_index]
    
    Parameters
    ----------
    left : Sequence
        first sequence of values
    right : Sequence
        second sequence of values
        
    Returns
    -------
    Tuple[MatchIndexes]
        tuple of namedtuple[value, left_index, right_index]
    
    Example
    -------
    >>> l1 = ['a', 'c', 'd', 'f', 'a']
    >>> l2 = ['a', 'b', 'c', 'd', 'c']
    >>> inner_matching_indexes(l1, l2)
    (MatchIndexes(value='a', left_index=0, right_index=0),
     MatchIndexes(value='c', left_index=1, right_index=2),
     MatchIndexes(value='c', left_index=1, right_index=4),
     MatchIndexes(value='d', left_index=2, right_index=3),
     MatchIndexes(value='a', left_index=4, right_index=0))
    """
    out = []
    for i, value in enumerate(left):
        right_i = locate(right, value)
        count = len(right_i)
        out.extend(_name_matches(zip(repeat(value, count), repeat(i, count), right_i)))
    return tuple(out)


def left_matching_indexes(
    left: Sequence,
    right: Sequence
) -> Tuple[MatchIndexes]:
    """
    Find matching item value indexes in two sequences.
    Returns tuple of tuple[value, left_index, right_index].
    Also, provide (value, left_index, None) pairs for unmatched values in left.
    
    Parameters
    ----------
    left : Sequence
    right : Sequence
    
    Returns
    -------
    Tuple[namedtuple[value, left_index, right_index]]
    
    Example
    -------
    >>> l1 = ['a', 'c', 'd', 'f', 'a', 'g']
    >>> l2 = ['a', 'b', 'c', 'd', 'c']
    >>> left_matching_indexes(l1, l2)
    (MatchIndexes(value='a', left_index=0, right_index=0),
     MatchIndexes(value='c', left_index=1, right_index=2),
     MatchIndexes(value='c', left_index=1, right_index=4),
     MatchIndexes(value='d', left_index=2, right_index=3),
     MatchIndexes(value='f', left_index=3, right_index=None),
     MatchIndexes(value='a', left_index=4, right_index=0),
     MatchIndexes(value='g', left_index=5, right_index=None))
    """
    out = []
    for i, value in enumerate(left):
        right_i = locate(right, value)
        if right_i:
            count = len(right_i)
            out.extend(_name_matches(zip(repeat(value, count), repeat(i, count), right_i)))
        else:
            out.append(MatchIndexes(value, left_index=i))
    return tuple(out)


def right_matching_indexes(
    l1: Sequence,
    l2: Sequence
) -> Tuple[MatchIndexes]:
    """
    Find matching item value indexes in two lists.
    Also, provide (value, None, right_index) pairs for unmatched values in right.
    
    Parameters
    ----------
    left : Sequence
    right : Sequence
    
    Returns
    -------
    Tuple[namedtuple[value, left_index, right_index]]
    
    Example
    -------
    >>> l1 = ['a', 'c', 'd', 'f', 'a', 'g']
    >>> l2 = ['a', 'b', 'c', 'd', 'c']
    >>> right_matching_indexes(l1, l2)
    (MatchIndexes(value='a', left_index=0, right_index=0),
     MatchIndexes(value='a', left_index=4, right_index=0),
     MatchIndexes(value='b', left_index=None, right_index=1),
     MatchIndexes(value='c', left_index=1, right_index=2),
     MatchIndexes(value='d', left_index=2, right_index=3),
     MatchIndexes(value='c', left_index=1, right_index=4))
    """
    out = []
    for i, value in enumerate(l2):
        l1_i = locate(l1, value)
        if l1_i:
            count = len(l1_i)
            out.extend(_name_matches(zip(repeat(value, count), l1_i, repeat(i, count))))
        else:
            out.append(MatchIndexes(value, right_index=i))
    return tuple(out)


def full_matching_indexes(
    left: Sequence,
    right: Sequence
) -> Tuple[MatchIndexes]:
    """
    Find matching item value indexes in two lists.
    Also, provide (value, left_index, None) pairs for unmatched values in left.
    Also, provide (value, None, right_index) pairs for unmatched values in right.
    
    Parameters
    ----------
    left : Sequence
    right : Sequence
    
    Returns
    -------
    Tuple[namedtuple[value, left_index, right_index]]
    
    Example
    -------
    >>> l1 = ['a', 'c', 'd', 'f', 'a', 'g']
    >>> l2 = ['a', 'b', 'c', 'd', 'c']
    >>> matching_indexes(l1, l2)
    (MatchIndexes(value='a', left_index=0, right_index=0),
     MatchIndexes(value='c', left_index=1, right_index=2),
     MatchIndexes(value='c', left_index=1, right_index=4),
     MatchIndexes(value='d', left_index=2, right_index=3),
     MatchIndexes(value='f', left_index=3, right_index=None),
     MatchIndexes(value='a', left_index=4, right_index=0),
     MatchIndexes(value='g', left_index=5, right_index=None),
     MatchIndexes(value='b', left_index=None, right_index=1))
    """
    out = []
    for i, value in enumerate(left):
        right_i = locate(right, value)
        if right_i:
            count = len(right_i)
            out.extend(_name_matches(zip(repeat(value, count), repeat(i, count), right_i)))
        else:
            out.append(MatchIndexes(value, left_index=i))    
    found_rights = set(x.right_index for x in out)
    unfound_rights = [MatchIndexes(value, right_index=i) for i, value in enumerate(right)
                          if i not in found_rights]
    out.extend(unfound_rights)
    return tuple(out)


def filter_values_by_index_matches(
    values: Sequence,
    indexes: Sequence[Union[int, None]]
) -> list:
    """
    Filter a sequence by indexes.
    Returns a list of matched index values and Nones for Nones in indexes.
    
    Parameters
    ----------
    values : Sequence
        sequence of values
    indexes : Sequence[Union[int, None]]
         sequence of indexes or Nones
         
    Returns
    -------
    list
        list of values or Nones at given indexes
    
    Example
    -------
    >>> values = ['x', 'y', 'z']
    >>> indexes = [0, None, 1, 2, None, 1]
    >>> filter_values_by_index_matches(values, indexes)
    ['x', None, 'y', 'z', None, 'y']
    """
    return [None if i is None else values[i] for i in indexes]


def _missing_col_error(col: str, table_name: str) -> ValueError:
    return ValueError(f'column {col} is missing from {table_name} table')


def _sequence_of_str(value) -> bool:
    if len(value) == 0:
        return False
    if isinstance(value, str):
        return False
    return all(isinstance(x, str) for x in value)


def _check_on_types(left_on, right_on) -> None:
    error = ValueError('right_on and left_on must both be str or sequence of str.')
    right_is_str_sequence = _sequence_of_str(right_on)
    left_is_str = isinstance(left_on, str)
    left_is_str_sequence = _sequence_of_str(left_on)
    right_is_str = isinstance(right_on, str)
    if not ((right_is_str and left_is_str) or (right_is_str_sequence and left_is_str_sequence)):
        raise error
    if left_is_str_sequence and right_is_str_sequence:
        if len(left_on) != len(right_on):
            raise ValueError('left_on sequence must be same len as right_on sequence')


def _check_for_missing_on(
    table: Mapping[str, Sequence],
    on_name: Union[str, Sequence[str]],
    table_name: str
) -> None:
    if isinstance(on_name, str):
        if on_name not in table:
            raise _missing_col_error(on_name, table_name)
    else:
        for col in on_name:
            if col not in table:
                _missing_col_error(col, table_name)


def tuple_keys(
    table: DataMapping,
    column_names: Sequence[str]
) -> Tuple[Tuple]:
    """
    Return tuple row values for just column_names.
    
    Parameters
    ----------
    table : DataMapping
        data mapping of {column name: column values}
    column_names : Sequence[str]
        column names to include in row values tuples
        
    Returns
    -------
    tuple[tuple]
        row values tuples
        
    Example
    -------
    >>> table = {'id': [1, 2, 3, 4],
                 'x': [3, 4, 2, 1],
                 'y': [4, 3, 2, 1]}
    >>> tuple_keys(table, ['x', 'y'])
    ((3, 4), (4, 3), (2, 2), (1, 1))
    """
    return tuple(zip(*[table[col] for col in column_names]))


def join(
    left: DataMapping,
    right: DataMapping,
    left_on: str,
    right_on: Optional[str] = None,
    select: Optional[Sequence[str]] = None,
    join_strategy: JoinStrategy = full_matching_indexes
) -> DataDict:
    """
    Join two data mappings on a specified column name(s)
    using a join strategy (inner, left, right or full).
    Default join strategy is full outer join if no
    join_strategy is passed.
    
    Parameters
    ----------
    left : Mapping[str, Sequence]
        left data mapping of {column name: column values}
    right : Mapping[str, Sequence]
        right data mapping of {column name: column values}
    left_on : str
        column name to join on in left
    right_on : str, optional
        column name to join on in right, join on left_on if None
    select : list[str], optional
        column names to return
    join_strategy : Callable[[Sequence, Sequence], Matches]
    
    Returns
    -------
    DataDict
        resulting joined data table
    
    Examples
    --------
    >>> left = {'id': ['a', 'c', 'd', 'f', 'g'], 'x': [33, 44, 55, 66, 77]}
    >>> right = {'id': ['a', 'b', 'c', 'd'], 'y': [11, 22, 33, 44]}
    >>> join(left, right, 'id', join_strategy=inner_matching_indexes)
    {'id': ['a', 'c', 'd'], 'x': [33, 44, 55], 'y': [11, 33, 44]}
    
    >>> left = {'id': ['a', 'c', 'd', 'f', 'g'], 'x': [33, 44, 55, 66, 77]}
    >>> right = {'id': ['a', 'b', 'c', 'd'], 'y': [11, 22, 33, 44]}
    >>> join(left, right, 'id', join_strategy=left_matching_indexes)
    {'id': ['a', 'c', 'd', 'f', 'g'],
     'x': [33, 44, 55, 66, 77],
     'y': [11, 33, 44, None, None]}
    
    >>> left = {'id': ['a', 'c', 'd', 'f', 'g'], 'x': [33, 44, 55, 66, 77]}
    >>> right = {'id': ['a', 'b', 'c', 'd'], 'y': [11, 22, 33, 44]}
    >>> join(left, right, 'id')
    {'id': ['a', 'c', 'd', 'f', 'g', 'b'],
     'x': [33, 44, 55, 66, 77, None],
     'y': [11, 33, 44, None, None, 22]}
    """
    right_on = left_on if right_on is None else right_on
    _check_on_types(left_on, right_on)
    _check_for_missing_on(left, left_on, 'left')
    _check_for_missing_on(right, right_on, 'right')
    
    if isinstance(left_on, str):
        indexes = join_strategy(left[left_on], right[right_on])
    else:
        indexes = join_strategy(tuple_keys(left, left_on), tuple_keys(right, right_on))

    values = [x.value for x in indexes]
    left_indexes = [x.left_index for x in indexes]
    right_indexes = [x.right_index for x in indexes]
    out = {col: filter_values_by_index_matches(left[col], left_indexes) for col in left}
    
    for col in right:
        if col not in [left_on, right_on]:
            out[col] = filter_values_by_index_matches(right[col], right_indexes)
            
    if isinstance(right_on, str):
        out[right_on] = values
        out[left_on] = values
    else:
        for i, col in enumerate(right_on):
            out[col] = [x[i] for x in values]
        for i, col in enumerate(left_on):
            out[col] = [x[i] for x in values]
    if select:
        return {col: out[col] for col in select}
    return out


def inner_join(
    left: DataMapping,
    right: DataMapping,
    left_on: str,
    right_on: Optional[str] = None,
    select: Optional[Sequence[str]] = None
) -> DataDict:
    """
    Inner Join two data dict on a specified column name(s).
    If right_on is None, joins both on same column name (left_on).
    Parameters
    ----------
    left : Mapping[str, Sequence]
        left data mapping of {column name: column values}
    right : Mapping[str, Sequence]
        right data mapping of {column name: column values}
    left_on : str
        column name to join on in left
    right_on : str, optional
        column name to join on in right, join on left_on if None
    select : list[str], optional
        column names to return
        
    Returns
    -------
    DataDict
        resulting joined data table
        
    Example
    -------
    >>> left =  {'id': ['a', 'c', 'd', 'f', 'g'], 'x': [33, 44, 55, 66, 77]}
    >>> right = {'id': ['a', 'b', 'c', 'd'], 'y': [11, 22, 33, 44]}
    >>> inner_join(left, right, 'id')
    {'id': ['a', 'c', 'd'], 'x': [33, 44, 55], 'y': [11, 33, 44]}
    """
    return join(left, right, left_on, right_on, select, inner_matching_indexes)


def full_join(
    left: DataMapping,
    right: DataMapping,
    left_on: str,
    right_on: Optional[str] = None,
    select: Optional[Sequence[str]] = None
) -> DataDict:
    """
    Full Join two data dict on a specified column name(s).
    If right_on is None, joins both on same column name (left_on).
    Parameters
    ----------
    left : Mapping[str, Sequence]
        left data mapping of {column name: column values}
    right : Mapping[str, Sequence]
        right data mapping of {column name: column values}
    left_on : str
        column name to join on in left
    right_on : str, optional
        column name to join on in right, join on left_on if None
    select : list[str], optional
        column names to return
        
    Returns
    -------
    DataDict
        resulting joined data table
        
    Example
    -------
    >>> left = {'id': ['a', 'c', 'd', 'f', 'g'], 'x': [33, 44, 55, 66, 77]}
    >>> right = {'id': ['a', 'b', 'c', 'd'], 'y': [11, 22, 33, 44]}
    >>> full_join(left, right, 'id')
    {'id': ['a', 'c', 'd', 'f', 'g', 'b'],
     'x': [33, 44, 55, 66, 77, None],
     'y': [11, 33, 44, None, None, 22]}
    """
    return join(left, right, left_on, right_on, select, full_matching_indexes)


def left_join(
    left: dict,
    right: dict,
    left_on: str,
    right_on: Optional[str] = None,
    select: Optional[Sequence[str]] = None
) -> dict:
    """
    Left Join two data dict on a specified column name(s).
    If right_on is None, joins both on same column name (left_on).
    
    Parameters
    ----------
    left : Mapping[str, Sequence]
        left data mapping of {column name: column values}
    right : Mapping[str, Sequence]
        right data mapping of {column name: column values}
    left_on : str
        column name to join on in left
    right_on : str, optional
        column name to join on in right, join on left_on if None
    select : list[str], optional
        column names to return
        
    Returns
    -------
    DataDict
        resulting joined data table
        
    Example
    -------
    >>> left = {'id': ['a', 'c', 'd', 'f'], 'x': [33, 44, 55, 66]}
    >>> right = {'id': ['a', 'b', 'c', 'd'], 'y': [11, 22, 33, 44]}
    >>> left_join(left, right, 'id')
    {'id': ['a', 'c', 'd', 'f'], 'x': [33, 44, 55, 66], 'y': [11, 33, 44, None]}
    """
    return join(left, right, left_on, right_on, select, left_matching_indexes)


def right_join(
    left: dict,
    right: dict,
    left_on: str,
    right_on: Optional[str] = None,
    select: Optional[Sequence[str]] = None
) -> dict:
    """
    Right Join two data dict on a specified column name(s).
    If right_on is None, joins both on same column name (left_on).
    
    Parameters
    ----------
    left : Mapping[str, Sequence]
        left data mapping of {column name: column values}
    right : Mapping[str, Sequence]
        right data mapping of {column name: column values}
    left_on : str
        column name to join on in left
    right_on : str, optional
        column name to join on in right, join on left_on if None
    select : list[str], optional
        column names to return
        
    Returns
    -------
    DataDict
        resulting joined data table
        
    Example
    -------
    >>> left = {'id': ['a', 'c', 'd', 'f'], 'x': [33, 44, 55, 66]}
    >>> right = {'id': ['a', 'b', 'c', 'd'], 'y': [11, 22, 33, 44]}
    >>> right_join(left, right, 'id')
    {'id': ['a', 'b', 'c', 'd'], 'x': [33, None, 44, 55], 'y': [11, 22, 33, 44]}
    """
    return join(left, right, left_on, right_on, select, right_matching_indexes)


from typing import Any, Dict, List, Mapping, MutableMapping, MutableSequence, Sequence, Union

import tinytim.data as data_features
import tinytim.copy as copy

MutableDataMapping = MutableMapping[str, MutableSequence]


def edit_row_items_inplace(
    data: MutableDataMapping,
    index: int,
    items: Mapping[str, Any]
) -> None:
    """
    Changes row index to mapping items values.

    Parameters
    ----------
    data : MutableMapping[str, MutableSequence]
        data mapping of {column name: column values}
    index : int
        index of row to edit
    items : Mapping[str, Any]
        {column names: value} of new values to edit in data

    Returns
    -------
    None

    Example
    -------
    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> edit_row_items_inplace(data, 0, {'x': 11, 'y': 66})
    >>> data
    {'x': [11, 2, 3], 'y': [66, 7, 8]}
    """
    for col in items:
        data[col][index] = items[col]


def edit_row_values_inplace(
    data: MutableDataMapping,
    index: int,
    values: Sequence
) -> None:
    """
    Changed row index to values.

    Parameters
    ----------
    data : MutableMapping[str, MutableSequence]
        data mapping of {column name: column values}
    index : int
        index of row to edit
    values : Sequence
        new values to replace in data row

    Returns
    -------
    None

    Example
    -------
    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> edit_row_values_inplace(data, 1, (22, 77))
    >>> data
    {'x': [1, 22, 3], 'y': [6, 77, 8]}
    """
    if len(values) != data_features.column_count(data):
        raise AttributeError('values length must match columns length.')
    for col, value in zip(data_features.column_names(data), values):
        data[col][index] = value


def edit_column_inplace(
    data: MutableDataMapping,
    column_name: str,
    values: MutableSequence
) -> None:
    """
    Add values to data in named column.
    Overrides existing values if column exists,
    Created new column with values if column does not exist.

    Parameters
    ----------
    data : MutableMapping[str, MutableSequence]
        data mapping of {column name: column values}
    column_name : str
        column name to edit in data
    values : Sequence
        new values to replace in data column

    Returns
    -------
    None

    Example
    -------
    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> edit_column_inplace(data, 'x', [11, 22, 33])
    >>> data
    {'x': [11, 22, 33], 'y': [6, 7, 8]}
    """
    if len(values) != data_features.row_count(data):
        raise ValueError('values length must match data rows count.')
    data[column_name] = values


def drop_row_inplace(
    data: MutableDataMapping,
    index: int
) -> None:
    """
    Remove index row from data.

    Parameters
    ----------
    data : MutableMapping[str, MutableSequence]
        data mapping of {column name: column values}
    index : int
        index of row to remove from data

    Returns
    -------
    None

    Example
    -------
    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> drop_row_inplace(data, 1)
    >>> data
    {'x': [1, 3], 'y': [6, 8]}
    """
    for col in data_features.column_names(data):
        data[col].pop(index)


def drop_label_inplace(labels: Union[None, List], index) -> None:
    """
    If labels exists, drop item at index.

    Parameters
    ----------
    labels : list, optional
        list of values used as labels
    index : int
        index of value to remove from labels list

    Returns
    -------
    None

    Examples
    --------
    >>> labels = [1, 2, 3, 4, 5]
    >>> drop_label_inplace(labels, 1)
    >>> labels
    [1, 3, 4, 5]

    >>> labels = None
    >>> drop_label_inplace(labels, 1)
    >>> labels
    None
    """
    if labels is not None:
        labels.pop(index)


def drop_column_inplace(
    data: MutableDataMapping,
    column_name: str
) -> None:
    """
    Remove named column from data.

    Parameters
    ----------
    data : MutableMapping[str, MutableSequence]
        data mapping of {column name: column values}
    column_name : str
        name of column to remove from data

    Returns
    -------
    None

    Example
    -------
    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> drop_column_inplace(data, 'y')
    >>> data
    {'x': [1, 2, 3]}
    """
    del data[column_name]


def edit_value_inplace(
    data: MutableDataMapping,
    column_name: str,
    index: int,
    value: Any
) -> None:
    """
    Edit the value in named column as row index.

    Parameters
    ----------
    data : MutableMapping[str, MutableSequence]
        data mapping of {column name: column values}
    column_name : str
        name of column to remove from data
    index : int
        row index of column to edit
    value : Any
        new value to change to

    Returns
    -------
    None

    Example
    -------
    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> edit_value_inplace(data, 'x', 0, 11)
    >>> data
    {'x': [11, 2, 3], 'y': [6, 7, 8]}
    """
    data[column_name][index] = value


def replace_column_names(
    data: MutableDataMapping,
    new_names: Sequence[str]
) -> Dict[str, MutableSequence]:
    """
    Return a new dict same column data but new column names.

    Parameters
    ----------
    data : MutableMapping[str, MutableSequence]
        data mapping of {column name: column values}
    new_names : Sequence[str]
        new names of columns

    Returns
    -------
    Dict[str, MutableSequence]
        copy of data with new column names

    Example
    -------
    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> replace_column_names(data, ('xx', 'yy'))
    >>> {'xx': [1, 2, 3], 'yy': [6, 7, 8]}
    >>> data
    {'x': [1, 2, 3], 'y': [6, 7, 8]}
    """
    old_names = data_features.column_names(data)
    if len(new_names) != len(old_names):
        raise ValueError('new_names must be same size as data column_count.')
    return {new_name: data[old_name] for new_name, old_name in zip(new_names, old_names)}


def edit_row_items(
    data: MutableDataMapping,
    index: int,
    items: Mapping
) -> MutableDataMapping:
    """
    Return a new dict with row index changed to mapping items values.

    Parameters
    ----------
    data : MutableMapping[str, MutableSequence]
        data mapping of {column name: column values}
    index : int
        index of row to edit
    items : Mapping[str, Any]
        {column names: value} of new values to edit in data

    Returns
    -------
    MutableMapping[str, MutableSequence]
        copy of data with edited row values

    Examples
    --------
    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> edit_row_items(data, 2, {'x': 33, 'y': 88})
    {'x': [1, 2, 33], 'y': [6, 7, 88]}
    >>> data
    {'x': [1, 2, 3], 'y': [6, 7, 8]}

    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> edit_row_items(data, 0, {'x': 55})
    {'x': [55, 2, 3], 'y': [6, 7, 8]}
    >>> data
    {'x': [1, 2, 3], 'y': [6, 7, 8]}
    """
    new_data = copy.copy_table(data)
    edit_row_items_inplace(new_data, index, items)
    return new_data


def edit_row_values(
    data: MutableDataMapping,
    index: int,
    values: Sequence
) -> MutableDataMapping:
    """
    Return a new dict with row index changed to values.

    Parameters
    ----------
    data : MutableMapping[str, MutableSequence]
        data mapping of {column name: column values}
    index : int
        index of row to edit
    values : Sequence
        new values to replace in data row

    Returns
    -------
    MutableMapping[str, MutableSequence]
        copy of data with row at index changed to values

    Example
    -------
    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> edit_row_values(data, 1, (22, 77))
    {'x': [1, 22, 3], 'y': [6, 77, 8]}
    >>> data
    {'x': [1, 2, 3], 'y': [6, 7, 8]}
    """
    new_data = copy.copy_table(data)
    edit_row_values_inplace(new_data, index, values)
    return new_data


def edit_column(
    data: MutableDataMapping,
    column_name: str,
    values: MutableSequence
) -> MutableDataMapping:
    """
    Returns a new dict with values added to data in named column.
    Overrides existing values if column exists,
    Created new column with values if column does not exist.

    Parameters
    ----------
    data : MutableMapping[str, MutableSequence]
        data mapping of {column name: column values}
    column_name : str
        column name to edit in data
    values : Sequence
        new values to replace in data column

    Returns
    -------
    MutableMapping[str, MutableSequence]
        copy of data with new column values

    Example
    -------
    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> edit_column(data, 'x', [4, 5, 6])
    {'x': [4, 5, 6], 'y': [6, 7, 8]}
    >>> data
     {'x': [1, 2, 3], 'y': [6, 7, 8]}
    """
    new_data = copy.copy_table(data)
    edit_column_inplace(new_data, column_name, values)
    return new_data


def edit_value(
    data: MutableDataMapping,
    column_name: str,
    index: int,
    value: Any
) -> MutableDataMapping:
    """
    Return a new table with the value in named column changed at row index.

    Parameters
    ----------
    data : MutableMapping[str, MutableSequence]
        data mapping of {column name: column values}
    column_name : str
        name of column to remove from data
    index : int
        row index of column to edit
    value : Any
        new value to change to

    Returns
    -------
    MutableMapping[str, MutableSequence]
        copy of data with new value changed

    Example
    -------
    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> edit_value(data, 'y', 2, 88)
    {'x': [1, 2, 3], 'y': [6, 7, 88]}
    >>> data
    {'x': [1, 2, 3], 'y': [6, 7, 8]}
    """
    new_data = copy.copy_table(data)
    edit_value_inplace(new_data, column_name, index, value)
    return new_data


def drop_row(
    data: MutableDataMapping,
    index: int
) -> MutableDataMapping:
    """
    Return a new dict with index row removed from data.

    Parameters
    ----------
    data : MutableMapping[str, MutableSequence]
        data mapping of {column name: column values}
    index : int
        index of row to remove from data

    Returns
    -------
    MutableMapping[str, MutableSequence]
        copy of data with dropped row

    Example
    -------
    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> drop_row(data, 0)
    {'x': [2, 3], 'y': [7, 8]}
    >>> data
    {'x': [1, 2, 3], 'y': [6, 7, 8]}
    """
    new_data = copy.copy_table(data)
    drop_row_inplace(new_data, index)
    return new_data


def drop_label(
    labels: Union[None, List],
    index: int
) -> Union[None, List]:
    """
    If labels exists, drop item at index.

    Parameters
    ----------
    labels : list, optional
        list of values used as labels
    index : int
        index of value to remove from labels list

    Returns
    -------
    list | None
        copy of labels with row dropped

    Examples
    --------
    >>> labels = [1, 2, 3, 4]
    >>> drop_label(labels, 1)
    [1, 3, 4]
    >>> labels
    [1, 3, 4, 5]

    >>> labels = None
    >>> drop_label(labels, 1)
    None
    >>> labels
    None
    """
    if labels is None: return
    new_labels = copy.copy_list(labels)
    drop_label_inplace(new_labels, index)
    return new_labels


def drop_column(
    data: MutableDataMapping,
    column_name: str
) -> MutableDataMapping:
    """
    Return a new dict with the named column removed from data.

    Parameters
    ----------
    data : MutableMapping[str, MutableSequence]
        data mapping of {column name: column values}
    column_name : str
        name of column to remove from data

    Returns
    -------
    MutableMapping[str, MutableSequence]
        copy of data with column dropped

    Example
    -------
    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> drop_column(data, 'y')
    {'x': [1, 2, 3]}
    >>> data
    {'x': [1, 2, 3], 'y': [6, 7, 8]}
    """
    new_data = copy.copy_table(data)
    drop_column_inplace(new_data, column_name)
    return new_data


"""
Module used for converting data format to json and json to data format.
"""

from typing import Dict, List, Mapping, Sequence
import json

from tinytim.rows import iterrows
from tinytim.utils import row_dicts_to_data


def data_to_json_list(data: Mapping[str, Sequence]) -> List[Dict]:
    """
    Convert data table to list of row dicts.

    Parameters
    ----------
    data : Mapping[str, Sequence]
        data mapping of {column name: column values}

    Returns
    -------
    list[dict]
        list of row dicts

    Example
    -------
    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> data_to_json_list(data)
    [{'x': 1, 'y': 6}, {'x': 2, 'y': 7}, {'x': 3, 'y': 8}]
    """
    return [row for _, row in iterrows(data)]


def json_list_to_data(l: List[Dict]) -> Dict[str, list]:
    """
    Convert list of row dicts to data table format.

    Parameters
    ----------
    l : list[dict]
        list of row dicts

    Returns
    -------
    dict[str, list]
        data dict of {column name: column values}

    Example
    -------
    >>>json = [{'x': 1, 'y': 6}, {'x': 2, 'y': 7}, {'x': 3, 'y': 8}]
    >>> json_list_to_data(json)
    {'x': [1, 2, 3], 'y': [6, 7, 8]}
    """
    return row_dicts_to_data(l)


def data_to_json(data: Mapping[str, Sequence]) -> str:
    """
    Convert data table to list of row dicts json string.

    Parameters
    ----------
    data : Mapping[str, Sequence]
        data mapping of {column name: column values}

    Returns
    -------
    str
        json string, list of row dicts

    Example
    -------
    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> data_to_json(data)
    '[{"x": 1, "y": 6}, {"x": 2, "y": 7}, {"x": 3, "y": 8}]'
    """
    l: List[Dict] = data_to_json_list(data)
    return json.dumps(l)


def json_to_data(j: str) -> Dict[str, list]:
    """
    Convert row dicts json string to data dict table.

    Parameters
    ----------
    j : str
        json string, list of row dicts

    Returns
    -------
    Mapping[str, Sequence]
        data mapping of {column name: column values}

    Example
    -------
    >>> j = '[{"x": 1, "y": 6}, {"x": 2, "y": 7}, {"x": 3, "y": 8}]'
    >>> json_to_data(j)
    {'x': [1, 2, 3], 'y': [6, 7, 8]}
    """
    l = json.loads(j)
    return json_list_to_data(l)


import copy
from typing import List, MutableMapping, MutableSequence

MutableDataMapping = MutableMapping[str, MutableSequence]


def copy_table(data: MutableDataMapping) -> MutableDataMapping:
    """
    Copy data and return the copy.

    Parameters
    ----------
    data : MutableMapping[str, MutableSequence]
        data mapping of {column name: column values}

    Returns
    -------
    MutableMapping[str, MutableSequence]
        copy of data

    Example
    -------
    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> copy_table(data)
    {'x': [1, 2, 3], 'y': [6, 7, 8]}
    """
    return copy.deepcopy(data)


def deepcopy_table(data: MutableDataMapping) -> dict:
    """
    Deep copy data and return the copy.

    Parameters
    ----------
    data : MutableMapping[str, MutableSequence]
        data mapping of {column name: column values}

    Returns
    -------
    MutableMapping[str, MutableSequence]
        copy of data

    Example
    -------
    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> deepcopy_table(data)
    {'x': [1, 2, 3], 'y': [6, 7, 8]}
    """
    return {col: copy.deepcopy(values) for col, values in data.items()}


def copy_list(values: List) -> List:
    """
    Copy list and return the copy.

    Parameters
    ----------
    values : list
        list of values

    Returns
    -------
    list
        copy of list

    Example
    -------
    >>> values = [1, 2, 3, 6, 7, 8]
    >>> values_copy = copy_list(values)
    >>> values_copy
    [1, 2, 3, 6, 7, 8]
    >>> values_copy[0] = 11
    >>> values_copy
    [11, 2, 3, 6, 7, 8]
    >>> values
    [1, 2, 3, 6, 7, 8]
    """
    return copy.copy(values)


def deepcopy_list(values: List) -> List:
    """
    Deep copy list and return the copy.

    Parameters
    ----------
    values : list
        list of values

    Returns
    -------
    list
        deep copy of list

    Example
    -------
    >>> values = [1, 2, 3, 6, 7, 8]
    >>> values_copy = deepcopy_list(values)
    >>> values_copy
    [1, 2, 3, 6, 7, 8]
    >>> values_copy[0] = 11
    >>> values_copy
    [11, 2, 3, 6, 7, 8]
    >>> values
    [1, 2, 3, 6, 7, 8]
    """
    return copy.deepcopy(values)


__version__ = '1.5.0'


import tinytim.columns
import tinytim.copy
import tinytim.data
import tinytim.edit
import tinytim.filter
import tinytim.group
import tinytim.rows
import tinytim.utils
import tinytim.validate
import tinytim.join



import random
from typing import Any, Callable, Dict, List, Mapping, Optional, Sequence

import tinytim.data as data_features


TableFilter = Sequence[bool]
DataMapping = Mapping[str, Sequence]
DataDict = Dict[str, Sequence]


def column_filter(column: Sequence, func: Callable[[Any], bool]) -> List[bool]:
    return [func(item) for item in column]


def indexes_from_filter(f: TableFilter) -> List[int]:
    return [i for i, b in enumerate(f) if b]


def filter_list_by_indexes(values: Sequence, indexes: Sequence[int]) -> List:
    """Return only values in indexes."""
    return [values[i] for i in indexes]


def filter_by_indexes(data: DataMapping, indexes: Sequence[int]) -> DataDict:
    """Return only rows in indexes"""
    return {col: filter_list_by_indexes(values, indexes) for col, values in data.items()}


def filter_data(data: DataMapping, f: TableFilter) -> DataDict:
    indexes = indexes_from_filter(f)
    return filter_by_indexes(data, indexes)


def filter_by_column_func(
    data: DataMapping,
    column_name: str,
    func: Callable[[Any], bool]
) -> DataDict:
    """Return only rows of data where named column equals value."""
    indexes = [i for i, val in enumerate(data[column_name]) if func(val)]
    return filter_by_indexes(data, indexes)


def filter_by_column_eq(data: DataMapping, column_name: str, value) -> DataDict:
    """Return only rows of data where named column equals value."""
    return filter_by_column_func(data, column_name, lambda x: x == value)


def filter_by_column_ne(data: DataMapping, column_name: str, value) -> DataDict:
    """Return only rows of data where named column does not equal value."""
    return filter_by_column_func(data, column_name, lambda x: x != value)


def filter_by_column_gt(data: DataMapping, column_name: str, value) -> DataDict:
    """Return only rows of data where named column is greater than value."""
    return filter_by_column_func(data, column_name, lambda x: x > value)


def filter_by_column_lt(data: DataMapping, column_name: str, value) -> DataDict:
    """Return only rows of data where named column is less than value."""
    return filter_by_column_func(data, column_name, lambda x: x < value)


def filter_by_column_ge(data: DataMapping, column_name: str, value) -> DataDict:
    """Return only rows of data where named column is greater than or equal value."""
    return filter_by_column_func(data, column_name, lambda x: x >= value)


def filter_by_column_le(data: DataMapping, column_name: str, value) -> DataDict:
    """Return only rows of data where named column is less than or equal value."""
    return filter_by_column_func(data, column_name, lambda x: x <= value)


def filter_by_column_isin(data: DataMapping, column_name: str, values: Sequence) -> DataDict:
    """Return only rows of data where named column is in values."""
    return filter_by_column_func(data, column_name, lambda x: x in values)


def filter_by_column_notin(data: DataMapping, column_name: str, values: Sequence) -> DataDict:
    """Return only rows of data where named column is not in values."""
    return filter_by_column_func(data, column_name, lambda x: x not in values)


def sample(data: DataMapping, n: int, random_state: Optional[int] = None) -> DataDict:
    """return random sample of n rows"""
    if random_state is not None:
        random.seed(random_state)
    indexes = random.sample(range(data_features.row_count(data)), n)
    return filter_by_indexes(data, indexes)


def only_columns(data: DataMapping, column_names: Sequence[str]) -> DataDict:
    """Return new TableDict with only column_names."""
    return {col: data[col] for col in column_names}


def sample_indexes(data: DataMapping, n: int, random_state: Optional[int] = None) -> List[int]:
    """return random sample of n indexes"""
    if random_state is not None:
        random.seed(random_state)
    return random.sample(range(data_features.row_count(data)), n)


from collections import defaultdict
from typing import Any, Collection, Dict, Generator, Iterable, List, Mapping
from typing import Optional, Sequence, Tuple

DataMapping = Mapping[str, Sequence]
RowMapping = Mapping[str, Any]


def uniques(values: Iterable) -> List:
    """
    Return a list of the unique items in values.

    Parameters
    ----------
    values : iterable
        iterable of objects of any type

    Returns
    -------
    list
        a list of unique objects from values

    Example
    -------
    >>> values = [1, 1, 2, 4, 5, 2, 0, 6, 1]
    >>> uniques(values)
    [1, 2, 4, 5, 0, 6]
    """
    out = []
    for value in values:
        if value not in out:
            out.append(value)
    return out


def nuniques(values: Sequence) -> int:
    """
    Count up number of unique items in values.

    Parameters
    ----------
    values : iterable
        iterable of objects of any type

    Returns
    -------
    int
        count of unique objects in values

    Example
    -------
    >>> values = [1, 1, 2, 4, 5, 2, 0, 6, 1]
    >>> nuniques(values)
    6
    """
    return len(uniques(values))


def row_value_tuples(
    data: DataMapping,
    column_names: Sequence[str]
) -> Tuple[tuple]:
    """
    Return row value tuples for only columns in column_names.

    Parameters
    ----------
    data : Mapping[str, Sequence]
        data mapping of {column name: column values}
    column_names : Sequence[str]
        sequence of column names

    Returns
    -------
    tuple[tuple]
        row value tuples for only columns in column_names

    Example
    -------
    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8], 'z': [9, 10, 11]}
    >>> row_value_tuples(data, ['x', 'z'])
    ((1, 9), (2, 10), (3, 11))
    """
    return tuple(zip(*[data[col] for col in column_names]))


def all_keys(dicts: Sequence[Mapping]) -> List:
    """
    Return all the unique keys from a collection of dicts.

    Parameters
    ----------
    dicts : Sequence[Mapping]
        sequence of mappings

    Returns
    -------
    list
        list of unique keys from all dicts

    Example
    -------
    >>> dicts = [{'x': 1, 'y': 2}, {'x': 4, 'z': 7}]
    >>> all_keys(dicts)
    ['x', 'y', 'z']
    """
    keys = []
    for d in dicts:
        for key in d:
            if key not in keys:
                keys.append(key)
    return keys


def row_dicts_to_data(
    rows: Sequence[RowMapping],
    missing_value: Optional[Any] = None
) -> Dict[str, list]:
    """
    Convert a list of row dicts to dict[col_name: values] format.

    Parameters
    ----------
    rows : Sequence[Mapping[str, Any]]
        sequence of row mappings
    missing_value : Any, optional
        value to insert if column is missing values

    Returns
    -------
    dict[str, list]
        data table formatted: {column name: column values}

    Examples
    --------
    >>> rows = [{'x': 1, 'y': 20}, {'x': 2, 'y': 21}, {'x': 3, 'y': 22}]
    >>> row_dicts_to_data(rows)
    {'x': [1, 2, 3], 'y': [20, 21, 22]}

    >>> rows = [{'x': 1, 'y': 20}, {'x': 2}, {'x': 3, 'y': 22}]
    >>> row_dicts_to_data(rows)
    {'x': [1, 2, 3], 'y': [20, None, 22]}
    """
    keys = all_keys(rows)
    data = defaultdict(list)
    for row in rows:
        for col in keys:
            if col in row:
                data[col].append(row[col])
            else:
                data[col].append(missing_value)
    return dict(data)


def all_bool(values: Collection) -> bool:
    """
    Return if all items in values are bool or not.

    Parameters
    ----------
    values : Collection
        collection of any values

    Returns
    -------
    bool
        True, if all values are bool type
        False, if not all values are bool type

    Examples
    --------
    >>> values = [1, True, False, True]
    >>> all_bool(values)
    False
    
    >>> values = [True, True, False, True]
    >>> all_bool(values)
    True
    """
    return all(isinstance(item, bool) for item in values)


def has_mapping_attrs(obj: Any) -> bool:
    """
    Check if object has all Mapping attrs.

    Parameters
    ----------
    obj : Any
        any object to check for mapping attrs

    Returns
    -------
    bool
        True, if obj has all mapping attrs
        False, if obj does not have all mapping attrs

    Examples
    --------
    >>> obj = dict()
    >>> has_mapping_attrs(obj)
    True
    
    >>> obj = list()
    >>> has_mapping_attrs(obj)
    False
    """
    mapping_attrs = ['__getitem__', '__iter__', '__len__',
                     '__contains__', 'keys', 'items', 'values',
                     'get', '__eq__', '__ne__']
    return all(hasattr(obj, a) for a in mapping_attrs)


def row_values_generator(row: RowMapping) -> Generator[Any, None, None]:
    """
    Return a generator that yields values from a row dict.

    Parameters
    ----------
    row : Mapping[str, Any]
        mapping of row values {column name: row value}

    Returns
    -------
    generator
        generator of row values

    Example
    -------
    >>> row = {'x': 1, 'y': 4, 'z': 8}
    >>> generator = row_values_generator(row)
    >>> next(generator)
    1
    >>> next(generator)
    4
    >>> next(generator)
    8
    >>> next(generator)
    ...
    StopIteration
    """
    for key in row:
        yield row[key]


def slice_to_range(s: slice, stop: Optional[int] = None) -> range:
    """
    Convert an int:int:int slice object to a range object.
    Needs stop if s.stop is None since range is not allowed to have stop=None.

    Parameters
    ----------
    s : slice
        slice object
    stop : int, optional
        stop is needed for range but not slice
        pass in if slice is missing stop

    Returns
    -------
    range
        range object with corresponding start, stop, step

    Examples
    --------
    >>> s = slice(1, 4, 0)
    >>> slice_to_range(s)
    range(0, 3, 1)

    >>> s = slice(0, 3, 1)
    >>> slice_to_range(s)
    range(0, 3, 1)
    """
    step = 1 if s.step is None else s.step
    if step == 0:
        raise ValueError('step must not be zero')

    if step > 0:
        start = 0 if s.start is None else s.start
        stop = s.stop if s.stop is not None else stop
    else:
        start = stop if s.start is None else s.start
        if isinstance(start, int):
            start -= 1
        stop = -1 if s.stop is None else s.stop

        if start is None:
            raise ValueError('start cannot be None is range with negative step')

    if stop is None:
        raise ValueError('stop cannot be None in range')
    
    return range(start, stop, step)


def combine_names_rows(
    column_names: Sequence[str],
    rows: Sequence[Sequence]
) -> Dict[str, List]:
    """
    Convert a sequence of column names and a sequence of row values
    into dict[column_name: values] format.

    Parameters
    ----------
    column_names : Sequence[str]
        sequence of column names
    rows : Sequence[Sequence]
        sequence of different row values

    Returns
    -------
    dict[str, list]
        data table formatted {column name: column values}

    Example
    -------
    >>> column_names = ['x', 'y']
    >>> rows = ((1, 2), (4, 5), (8, 10))
    >>> combine_names_rows(column_names, rows)
    {'x': [1, 4, 8], 'y': [2, 5, 10]}
    """
    return dict(zip(column_names, map(list, zip(*rows))))


def nunique(data: DataMapping) -> Dict[str, int]:
    """
    Count number of distinct values in each column.
    Return dict with count of distinct values.

    Parameters
    ----------
    data : Mapping[str, Sequence]
        data mapping of {column name: column values}

    Returns
    -------
    dict[str, int]
        dict formatted {column name: unique value count}

    Example
    -------
    >>> data = {'x': [1, 2, 2], 'y': [6, 7, 8], 'z': [9, 9, 9]}
    >>> nunique(data)
    {'x': 2, 'y': 3, 'z': 1}
    """
    return {col: len(uniques(values)) for col, values in data.items()}


from typing import Any, Dict, Mapping, Sequence, Tuple

DataMapping = Mapping[str, Sequence]


def column_count(data: DataMapping) -> int:
    """
    Return the number of columns in data.

    Parameters
    ----------
    data : Mapping[str, Sequence]
        data mapping of {column name: column values}

    Returns
    -------
    int
        number of columns in data

    Example
    -------
    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> column_count(data)
    2
    """
    return len(data)


def first_column_name(data: DataMapping) -> str:
    """
    Return the name of the first column.
    Raises StopIteration if data has zero columns.

    Parameters
    ----------
    data : Mapping[str, Sequence]
        data mapping of {column name: column values}

    Returns
    -------
    str
        name of first column

    Raises
    ------
    StopIteration
        if data has zero columns

    Example
    -------
    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> first_column_name(data)
    'x'
    """
    return next(iter(data))


def row_count(data: DataMapping) -> int:
    """
    Return the number of rows in data.
    
    Parameters
    ----------
    data : Mapping[str, Sequence]
        data mapping of {column name: column values}

    Returns
    -------
    int
        number of rows in data

    Example
    -------
    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> row_count(data)
    3
    """
    if column_count(data) == 0: return 0
    return len(data[first_column_name(data)])


def shape(data: DataMapping) -> Tuple[int, int]:
    """
    Return data row count, column count tuple.

    Parameters
    ----------
    data : Mapping[str, Sequence]
        data mapping of {column name: column values}

    Returns
    -------
    tuple[int, int]
        (row count, column count)

    Example
    -------
    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> shape(data)
    (3, 2)
    """
    col_count = column_count(data)
    if col_count == 0: return 0, 0
    return row_count(data), col_count


def size(data: DataMapping) -> int:
    """
    Return data row count multiplied by column count.
        
    Parameters
    ----------
    data : Mapping[str, Sequence]
        data mapping of {column name: column values}

    Returns
    -------
    int
        row count * column count

    Example
    -------
    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> size(data)
    6
    """
    rows, columns = shape(data)
    return rows * columns


def column_names(data: DataMapping) -> Tuple[str]:
    """
    Return data column names.

    Parameters
    ----------
    data : Mapping[str, Sequence]
        data mapping of {column name: column values}

    Returns
    -------
    tuple[str]
        tuple of data column names

    Example
    -------
    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> column_names(data)
    ('x', 'y')
    """
    return tuple(data)


def head(data: DataMapping, n: int = 5) -> Dict[str, Sequence]:
    """
    Return the first n rows of data.

    Parameters
    ----------
    data : Mapping[str, Sequence]
        data mapping of {column name: column values}
    n : int, optional
        number of rows to return from top of data

    Returns
    -------
    dict[str, Sequence]
        {column name: top n column values}

    Example
    -------
    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> head(data, 2)
    {'x': [1, 2], 'y': [6, 7]}
    """
    return {col: values[:n] for col, values in data.items()}


def tail(data: DataMapping, n: int = 5) -> Dict[str, Sequence]:
    """
    Return the last n rows of data.

    Parameters
    ----------
    data : Mapping[str, Sequence]
        data mapping of {column name: column values}
    
    n : int, optional
        number of rows to return from bottom of data

    Returns
    -------
    dict
        {column name: bottom n column values}

    Example
    -------
    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> tail(data, 2)
    {'x': [2, 3], 'y': [7, 8]}
    """
    return {col: values[-n:] for col, values in data.items()}


def index(data: DataMapping) -> Tuple[int]:
    """
    Return tuple of data column indexes.

    Parameters
    ----------
    data : Mapping[str, Sequence]
        data mapping of {column name: column values}

    Returns
    -------
    tuple[int]
        tuple of row indexes

    Example
    -------
    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> index(data)
    (0, 1, 2)
    """
    return tuple(range(row_count(data)))


def table_value(data: DataMapping, column_name: str, index: int) -> Any:
    """
    Return one value from column at row index.

    Parameters
    ----------
    data : Mapping[str, Sequence]
        data mapping of {column name: column values}
    column_name : str
        name of column
    index : int
        index of column value

    Returns
    -------
    Any
        value of column at index

    Example
    -------
    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> table_value(data, 'x', 1)
    2
    """
    return data[column_name][index]


def column_values(data: DataMapping, column_name: str) -> Sequence:
    """
    Return all the values from one column.
    
    Parameters
    ----------
    data : Mapping[str, Sequence]
        data mapping of {column name: column values}
    column_name : str
        name of column

    Returns
    -------
    Sequence
        column values

    Example
    -------
    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> column_values(data, 'y')
    [6, 7, 8]
    """
    return data[column_name]


from typing import Any, Callable, Dict, List, Mapping, Sequence, Tuple, Union
from statistics import mean, mode, stdev, pstdev

from tinytim.filter import column_filter, filter_data
from tinytim.utils import nuniques, row_dicts_to_data, row_value_tuples, uniques

DataMapping = Mapping[str, Sequence]
DataDict = Dict[str, list]
RowMapping = Mapping[str, Any]
RowDict = Dict[str, Any]
RowNumDict = Dict[str, Union[int, float]]


def groupbycolumn(data: Mapping, column: Sequence) -> List[tuple]:
    keys = uniques(column)
    return [(k, filter_data(data, column_filter(column, lambda x: x == k)))
                for k in keys]


def groupbyone(data: Mapping, column_name: str) -> List[tuple]:
    return groupbycolumn(data, data[column_name])


def groupbymulti(data: Mapping, column_names: Sequence[str]) -> List[tuple]:
    return groupbycolumn(data, row_value_tuples(data, column_names))


def groupby(data: Mapping, by: Union[str, Sequence[str]]) -> List[tuple]:
    if isinstance(by, str):
        return groupbyone(data, by)
    else:
        return groupbymulti(data, by)


def aggregate_groups(
    groups: Sequence[Tuple[Any, DataMapping]],
    func: Callable[[DataMapping], RowMapping]
) -> Tuple[List, DataDict]:
    labels = []
    rows = []
    for key, data in groups:
        row = func(data)
        if len(row):
            labels.append(key)
            rows.append(row)
    return labels, row_dicts_to_data(rows)


def sum_groups(groups: List[tuple]) -> Tuple[List, DataDict]:
    return aggregate_groups(groups, sum_data)


def count_groups(groups: List[tuple]) -> Tuple[List, DataDict]:
    return aggregate_groups(groups, count_data)


def mean_groups(groups: List[tuple]) -> Tuple[List, DataDict]:
    return aggregate_groups(groups, mean_data)


def min_groups(groups: List[tuple]) -> Tuple[List, DataDict]:
    return aggregate_groups(groups, min_data)


def max_groups(groups: List[tuple]) -> Tuple[List, DataDict]:
    return aggregate_groups(groups, max_data)


def mode_groups(groups: List[tuple]) -> Tuple[List, DataDict]:
    return aggregate_groups(groups, mode_data)


def stdev_groups(groups: List[tuple]) -> Tuple[List, DataDict]:
    return aggregate_groups(groups, stdev_data)


def pstdev_groups(groups: List[tuple]) -> Tuple[List, DataDict]:
    return aggregate_groups(groups, pstdev_data)


def nunique_groups(groups: List[tuple]) -> Tuple[List, DataDict]:
    return aggregate_groups(groups, nunique_data)


def aggregate_data(data: Mapping, func: Callable) -> RowDict:
    out = {}
    for column_name in data.keys():
        try:
            col_sum = func(data[column_name])
        except TypeError:
            continue
        out[column_name] = col_sum
    return out


def sum_data(data: Mapping) -> RowNumDict:
    return aggregate_data(data, sum)


def count_data(data: Mapping) -> RowNumDict:
    return aggregate_data(data, len)


def mean_data(data: Mapping) -> RowNumDict:
    return aggregate_data(data, mean)


def min_data(data: Mapping) -> RowNumDict:
    return aggregate_data(data, min)


def max_data(data: Mapping) -> RowNumDict:
    return aggregate_data(data, max)


def mode_data(data: Mapping) -> RowNumDict:
    return aggregate_data(data, mode)


def stdev_data(data: Mapping) -> RowNumDict:
    return aggregate_data(data, stdev)


def pstdev_data(data: Mapping) -> RowNumDict:
    return aggregate_data(data, pstdev)


def nunique_data(data: Mapping) -> RowNumDict:
    return aggregate_data(data, nuniques)


from typing import Dict, Generator, Iterable, Mapping, Sequence, Tuple

import tinytim.data as data_features
from dictanykey import DefaultDictAnyKey, DictAnyKey

DataMapping = Mapping[str, Sequence]


def column_dict(data: DataMapping, col: str) -> Dict[str, Sequence]:
    """
    Return a dict of {col_name, col_values} from data.
        
    Parameters
    ----------
    data : Mapping[str, Sequence]
        data mapping of {column name: column values}
    col : str
        column name to pull out of data.

    Returns
    -------
    dict[str, Sequence]
        {column_name: column_values}

    Example
    -------
    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> column_dict(data, 'x')
    {'x': [1, 2, 3]}
    >>> column_dict(data, 'y')
    {'y': [6, 7, 8]}
    """
    return {col: data_features.column_values(data, col)}


def itercolumns(data: DataMapping) -> Generator[Tuple[str, tuple], None, None]:
    """
    Return a generator of tuple column name, column values.

    Parameters
    ----------
    data : Mapping[str, Sequence]
        data mapping of {column name: column values}
    
    Returns
    -------
    Generator[Tuple[str, tuple], None, None]
        generator that yields tuples(column_name, column_values)
        
    Example
    -------
    >>> data = 'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> cols = list(itercolumns(data))
    >>> cols[0]
    ('x', (1, 2, 3)) 
    >>> cols[1]
    ('y', (6, 7, 8))
    """
    for col in data_features.column_names(data):
        yield col, tuple(data_features.column_values(data, col))


def value_counts(
   values: Iterable,
   sort=True,
   ascending=True
) -> DictAnyKey:
    """
    Count up each value.
    Return a DictAnyKey[value] -> count
    Allows for unhashable values.

    Parameters
    ----------
    values :
        values to be counted up
    sort : default True, sort results by counts
    ascending: default True, sort highest to lowest


    Returns
    -------
    DictAnyKey[Any, int]
        {value: value_count}

    Example
    -------
    >>> values = [4, 1, 1, 4, 5, 1]
    >>> value_counts(values)
    DictAnyKey((1, 3), (4, 2), (5, 1))
    """
    d = DefaultDictAnyKey(int)
    for value in values:
        d[value] += 1
    if sort:
        return DictAnyKey(sorted(d.items(),  # type: ignore
                                 key=lambda item: item[1],  # type: ignore
                                 reverse=ascending))
    else:
        return DictAnyKey(d)


import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()

setup(
    name="tinytim",
    version="1.5.0",
    description="Pure Python data table functions.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/eddiethedean/tinytim",
    author="Odos Matthews",
    author_email="odosmatthews@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
    ],
    packages=find_packages(),
    include_package_data=True,
    python_requires='>=3.6',
    install_requires=['dictanykey>=0.0.2']
)


from collections import defaultdict
from typing import Dict, Generator, Mapping, MutableMapping, Sequence, Tuple

import tinytim.data as data_features


def row_dict(
    data: Mapping,
    index: int
) -> dict: 
    """
    Return one row from data at index.

    Parameters
    ----------
    data : Mapping[str, Sequence]
        data mapping of {column name: column values}
    index : int
        row index

    Returns
    -------
    dict
        one row from data at index

    Example
    -------
    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> row_dict(data, 1)
    {'x': 2, 'y': 7}
    """
    return {col: data_features.table_value(data, col, index)
                for col in data_features.column_names(data)}


def row_values(
    data: MutableMapping,
    index: int
) -> tuple:
    """
    Return a tuple of the values at row index.

    Parameters
    ----------
    data : Mapping[str, Sequence]
        data mapping of {column name: column values}
    index : int
        row index

    Returns
    -------
    tuple
        values at row index

    Example
    -------
    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> row_values(data, 0)
    (1, 6)
    """
    return tuple(values[index] for values in data.values())


def iterrows(
    data: Mapping
) -> Generator[Tuple[int, dict], None, None]:
    """
    Return a generator of tuple row index, row dict values.

    Parameters
    ----------
    data : Mapping[str, Sequence]
        data mapping of {column name: column values}

    Returns
    -------
    generator[tuple[int, dict]]
        generator of tuple (row index, row dict values)

    Example
    -------
    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> generator = iterrows(data)
    >>> next(generator)
    (0, {'x': 1, 'y': 6})
    >>> next(generator)
    (1, {'x': 2, 'y': 7})
    >>> next(generator)
    (2, {'x': 3, 'y': 8})
    >>> next(generator)
    ...
    StopIteration
    """
    for i in data_features.index(data):
        yield i, row_dict(data, i)


def itertuples(
    data: Mapping
) -> Generator[tuple, None, None]:
    """
    Return a generator of tuple row values.

    Parameters
    ----------
    data : Mapping[str, Sequence]
        data mapping of {column name: column values}

    Returns
    -------
    generator[tuple]
        generator of tuple row values

    Example
    -------
    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> generator = iterrows(data)
    >>> next(generator)
    (1, 6)
    >>> next(generator)
    (2, 7)
    >>> next(generator)
    (3, 8)
    >>> next(generator)
    ...
    StopIteration
    """
    for _, row in iterrows(data):
        yield tuple(row.values())


def itervalues(
    data: MutableMapping
) -> Generator[tuple, None, None]:
    """
    Return a generator of tuple row values.

    Parameters
    ----------
    data : Mapping[str, Sequence]
        data mapping of {column name: column values}

    Returns
    -------
    generator[tuple]
        generator of tuple row values

    Example
    -------
    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> generator = itervalues(data)
    >>> next(generator)
    (1, 6)
    >>> next(generator)
    (2, 7)
    >>> next(generator)
    (3, 8)
    >>> next(generator)
    ...
    StopIteration
    """
    return itertuples(data)


def values(
    data: MutableMapping
) -> Tuple[tuple]:
    """
    Return tuple of tuple row values.

    Parameters
    ----------
    data : Mapping[str, Sequence]
        data mapping of {column name: column values}

    Returns
    -------
    tuple[tuple]
        tuple of tuple row values

    Example
    -------
    >>> data = {'x': [1, 2, 3], 'y': [6, 7, 8]}
    >>> values(data)
    ((1, 6), (2, 7), (3, 8))
    """
    return tuple(itervalues(data))


def row_value_counts(
    data: Mapping[str, Sequence],
    sort=True,
    ascending=True
) -> Dict[tuple, int]:
    """
    Count up the unique rows.

    Parameters
    ----------
    data : Mapping[str, Sequence]
        data mapping of {column name: column values}
    sort : bool, optional
        sort the results by count
    ascending : bool, optional
        if sort=True, sort highest to lowest

    Returns
    -------
    dict[tuple, int]
        {(row values), count}

    Example
    -------
    >>> data = {'x': [1, 2, 3, 3], 'y': [6, 7, 3, 3]}
    >>> row_value_counts(data)
    {(3, 3): 2, (1, 6): 1, (2, 7): 1}
    """
    d = defaultdict(int)
    for row in itertuples(data):
        d[row] += 1
    if sort:
        return dict(sorted(d.items(),
                           key=lambda item: item[1],
                           reverse=ascending))
    else:
        return dict(d)


def records(d: Mapping[str, Sequence]) -> Generator[Mapping, None, None]:
    """
    Yield each record (row) in d.
    
    Parameters
    ----------
    d : Mapping[str, Sequence]
        data mapping of {column name: column values}
    
    Example
    -------
    >>> d = {'x': [1, 2, 3, 4], 'y': [55, 66, 77, 88]}
    >>> generator = records(d)
    >>> next(generator)
    {'x': 1, 'y': 55}
    >>> next(generator)
    {'x': 2, 'y': 66}
    >>> next(generator)
    {'x': 3, 'y': 77}
    >>> next(generator)
    {'x': 4, 'y': 88}
    """
    for _, record in iterrows(d):
        yield record
    

def records_equal(d1: Mapping[str, Sequence], d2: Mapping[str, Sequence]) -> bool:
    """
    Compare d1 and d2 records (rows) to see if they are equal.
    Order of records or columns does not matter.
    
    Parameters
    ----------
    d1 : Mapping[str, Sequence]
        data mapping of {column name: column values}
    d2 : Mapping[str, Sequence]
        data mapping of {column name: column values}
    
    Examples
    --------
    >>> d1 = {'x': [1, 2, 3, 4], 'y': [55, 66, 77, 88]}
    >>> d2 = {'x': [2, 1, 4, 3], 'y': [66, 55, 88, 77]}
    >>> records_equal(d1, d2)
    True
    
    >>> d1 = {'x': [1, 2, 3, 4], 'y': [55, 66, 77, 88]}
    >>> d2 = {'x': [2, 1, 4, 3], 'y': [55, 77, 88, 66]}
    >>> records_equal(d1, d2)
    False
    """
    if set(data_features.column_names(d1)) != set(data_features.column_names(d2)):
        return False
    
    if data_features.row_count(d1) != data_features.row_count(d2):
        return False
    
    d2_rows = [row for row in records(d2)]
    
    for row in records(d1):
        if row in d2_rows:
            d2_rows.remove(row)
        else:
            return False
    
    return True